import asyncio
import click
import types
import typing as t
import watchfiles
import httpx
import subprocess
import sys
import time
import functools
import tomlkit
from pathlib import Path

from . import Agent, config, loader, decorators, models

T = t.TypeVar("T")


def _callback(_changes: set[tuple[watchfiles.Change, str]]) -> None:
    print("Change detected. Reloading...")


def _api_request(method: str, host: str, action: str, **kwargs) -> t.Any:
    with httpx.Client() as client:
        response = client.request(method, f"http://{host}/api/{action}", **kwargs)
        # TODO: return errors
        response.raise_for_status()
        is_json = response.headers.get("Content-Type") == "application/json"
        return response.json() if is_json else None


def _encode_provides(
    provides: dict[str, list[str] | str | bool] | None
) -> tuple[str, ...] | None:
    if not provides:
        return None
    return tuple(
        f"{k}:{str(v).lower() if isinstance(v, bool) else v}"
        for k, vs in provides.items()
        for v in (vs if isinstance(vs, list) else [vs])
    )


def _parse_provides(argument: tuple[str] | None) -> dict[str, list[str]]:
    if not argument:
        return {}
    result: dict[str, list[str]] = {}
    for part in (p for a in argument for p in a.split(" ") if p):
        key, value = part.split(":", 1)
        result.setdefault(key, []).append(value)
    return result


def _load_repository(
    module: types.ModuleType,
) -> dict[str, tuple[models.Target, t.Callable]]:
    attrs = (getattr(module, k) for k in dir(module))
    return {
        a.name: (a.definition, a.fn)
        for a in attrs
        if isinstance(a, decorators.Target) and not a.definition.is_stub
    }


def _load_repositories(
    modules: list[types.ModuleType | str],
) -> dict[str, dict[str, tuple[models.Target, t.Callable]]]:
    targets = {}
    for module in list(modules):
        if isinstance(module, str):
            module = loader.load_module(module)
        targets[module.__name__] = _load_repository(module)
    return targets


def _register_manifests(
    project_id: str,
    environment_name: str,
    host: str,
    targets: dict[str, dict[str, tuple[models.Target, t.Callable]]],
) -> None:
    manifests = {
        repository: {
            "workflows": {
                workflow_name: {
                    "parameters": [
                        {
                            "name": p.name,
                            "annotation": p.annotation,
                            "default": p.default,
                        }
                        for p in definition.parameters
                    ],
                    "waitFor": list(definition.wait_for),
                    "cache": (
                        definition.cache
                        and {
                            "params": definition.cache.params,
                            "maxAge": definition.cache.max_age,
                            "namespace": definition.cache.namespace,
                            "version": definition.cache.version,
                        }
                    ),
                    "defer": (
                        definition.defer
                        and {
                            "params": definition.defer.params,
                        }
                    ),
                    "delay": definition.delay,
                    "retries": (
                        definition.retries
                        and {
                            "limit": definition.retries.limit,
                            "delayMin": definition.retries.delay_min,
                            "delayMax": definition.retries.delay_max,
                        }
                    ),
                    "requires": definition.requires,
                    "instruction": definition.instruction,
                }
                for workflow_name, (definition, _) in target.items()
                if definition.type == "workflow"
            },
            "sensors": {
                sensor_name: {
                    "parameters": [
                        {
                            "name": p.name,
                            "annotation": p.annotation,
                            "default": p.default,
                        }
                        for p in definition.parameters
                    ],
                    "requires": definition.requires,
                    "instruction": definition.instruction,
                }
                for sensor_name, (definition, _) in target.items()
                if definition.type == "sensor"
            },
        }
        for repository, target in targets.items()
    }
    # TODO: handle response?
    _api_request(
        "POST",
        host,
        "register_manifests",
        json={
            "projectId": project_id,
            "environmentName": environment_name,
            "manifests": manifests,
        },
    )


def _init(
    *modules: types.ModuleType | str,
    project: str,
    environment: str,
    host: str,
    provides: dict[str, list[str]],
    serialiser_configs: list[config.SerialiserConfig],
    blob_threshold: int,
    blob_store_configs: list[config.BlobStoreConfig],
    concurrency: int,
    launch_id: str | None,
    register: bool,
) -> None:
    try:
        targets = _load_repositories(list(modules))
        if register:
            _register_manifests(project, environment, host, targets)

        with Agent(
            project,
            environment,
            host,
            provides,
            serialiser_configs,
            blob_threshold,
            blob_store_configs,
            concurrency,
            launch_id,
            targets,
        ) as agent:
            asyncio.run(agent.run())
    except KeyboardInterrupt:
        pass


@click.group()
def cli():
    pass


@cli.command("server")
@click.option(
    "-p",
    "--port",
    type=int,
    default=7777,
    help="Port to run server on",
)
@click.option(
    "-d",
    "--data-dir",
    type=click.Path(file_okay=False, path_type=Path, resolve_path=True),
    default="./data/",
    help="The directory to store data",
)
@click.option(
    "--image",
    default="ghcr.io/cofluxlabs/coflux",
    help="The Docker image to run",
)
def server(port: int, data_dir: Path, image: str):
    """
    Start a local server.

    This is just a wrapper around Docker (which must be installed and running), useful for running the server in a development environment.
    """
    command = [
        "docker",
        "run",
        "--pull",
        ("missing" if image.startswith("sha256:") else "always"),
        "--publish",
        f"{port}:7777",
        "--volume",
        f"{data_dir}:/data",
        image,
    ]
    process = subprocess.run(command)
    sys.exit(process.returncode)


def _config_path():
    return Path.cwd().joinpath("coflux.toml")


def _read_config(path: Path) -> tomlkit.TOMLDocument:
    if path.exists():
        with path.open("r") as f:
            return tomlkit.load(f)
    else:
        # TODO: add instructions?
        return tomlkit.document()


def _write_config(path: Path, data: tomlkit.TOMLDocument):
    with path.open("w") as f:
        tomlkit.dump(data, f)


@functools.cache
def _load_config() -> config.Config:
    path = _config_path()
    return config.Config.model_validate(_read_config(path).unwrap())


def _load_pools_config(file: t.TextIO) -> config.PoolsConfig:
    return config.PoolsConfig.model_validate(tomlkit.load(file).unwrap())


@cli.command("configure")
@click.option(
    "-p",
    "--project",
    help="Project ID",
    default=_load_config().project,
    show_default=True,
    prompt=True,
)
@click.option(
    "environment",
    "-e",
    "--environment",
    help="Environment name",
    default=_load_config().environment,
    show_default=True,
    prompt=True,
)
@click.option(
    "-h",
    "--host",
    help="Host to connect to",
    default=_load_config().server.host,
    show_default=True,
    prompt=True,
)
def configure(
    host: str | None,
    project: str | None,
    environment: str | None,
):
    """
    Populate/update the configuration file.
    """
    # TODO: connect to server to check details?
    click.secho("Writing configuration...", fg="black")

    path = _config_path()
    data = _read_config(path)
    data["project"] = project
    data["environment"] = environment
    data.setdefault("server", {})["host"] = host
    _write_config(path, data)

    click.secho(
        f"Configuration written to '{path.relative_to(Path.cwd())}'.", fg="green"
    )


@cli.group()
def env():
    """
    Manage environments.
    """
    pass


@env.command("create")
@click.option(
    "-p",
    "--project",
    help="Project ID",
    envvar="COFLUX_PROJECT",
    default=_load_config().project,
    show_default=True,
    required=True,
)
@click.option(
    "-h",
    "--host",
    help="Host to connect to",
    envvar="COFLUX_HOST",
    default=_load_config().server.host,
    show_default=True,
    required=True,
)
@click.option(
    "--base",
    help="The base environment to inherit from",
)
@click.option(
    "--pools-config",
    help="Path to pools configuration file",
    type=click.File(),
)
@click.argument("name")
def env_create(
    project: str,
    host: str,
    base: str | None,
    pools_config: t.TextIO,
    name: str,
):
    """
    Creates an environment within the project.
    """
    base_id = None
    if base:
        environments = _api_request(
            "GET", host, "get_environments", params={"project": project}
        )
        environment_ids_by_name = {e["name"]: id for id, e in environments.items()}
        base_id = environment_ids_by_name.get(base)
        if not base_id:
            click.BadOptionUsage("base", "Not recognised")

    pools = None
    if pools_config:
        pools = _load_pools_config(pools_config)

    # TODO: handle response
    _api_request(
        "POST",
        host,
        "create_environment",
        json={
            "projectId": project,
            "name": name,
            "baseId": base_id,
            "pools": pools.model_dump() if pools else None,
        },
    )
    click.secho(f"Created environment '{name}'.", fg="green")


@env.command("update")
@click.option(
    "-p",
    "--project",
    help="Project ID",
    envvar="COFLUX_PROJECT",
    default=_load_config().project,
    show_default=True,
    required=True,
)
@click.option(
    "-e",
    "--environment",
    help="The (current) name of the environment",
    envvar="COFLUX_ENVIRONMENT",
    default=_load_config().environment,
    show_default=True,
    required=True,
)
@click.option(
    "-h",
    "--host",
    help="Host to connect to",
    envvar="COFLUX_HOST",
    default=_load_config().server.host,
    show_default=True,
    required=True,
)
@click.option(
    "--name",
    help="The new name of the environment",
)
@click.option(
    "--base",
    help="The new base environment to inherit from",
)
@click.option(
    "--no-base",
    is_flag=True,
    help="Unset the base environment",
)
@click.option(
    "--pools-config",
    help="Path to pools configuration file",
    type=click.File(),
)
@click.option(
    "--no-pools",
    is_flag=True,
    help="Clear all pools from the environment",
)
def env_update(
    project: str,
    environment: str,
    host: str,
    name: str | None,
    base: str | None,
    no_base: bool,
    pools_config: t.TextIO,
    no_pools: bool,
):
    """
    Creates an environment within the project.
    """
    environments = _api_request(
        "GET", host, "get_environments", params={"project": project}
    )
    environment_ids_by_name = {e["name"]: id for id, e in environments.items()}
    environment_id = environment_ids_by_name.get(environment)
    if not environment_id:
        raise click.BadOptionUsage("environment", "Not recognised")

    base_id = None
    if base:
        base_id = environment_ids_by_name.get(base)
        if not base_id:
            raise click.BadOptionUsage("base", "Not recognised")

    pools = None
    if pools_config:
        pools = _load_pools_config(pools_config)

    payload = {
        "projectId": project,
        "environmentId": environment_id,
    }
    if name is not None:
        payload["name"] = name

    if base is not None:
        payload["baseId"] = base_id
    elif no_base is True:
        payload["baseId"] = None

    if pools is not None:
        payload["pools"] = pools.model_dump()
    elif no_pools is True:
        payload["pools"] = None

    # TODO: handle response
    _api_request("POST", host, "update_environment", json=payload)

    click.secho(f"Updated environment '{name or environment}'.", fg="green")


@env.command("archive")
@click.option(
    "-p",
    "--project",
    help="Project ID",
    envvar="COFLUX_PROJECT",
    default=_load_config().project,
    show_default=True,
    required=True,
)
@click.option(
    "-e",
    "--environment",
    help="Environment name",
    envvar="COFLUX_ENVIRONMENT",
    default=_load_config().environment,
    show_default=True,
    required=True,
)
@click.option(
    "-h",
    "--host",
    help="Host to connect to",
    envvar="COFLUX_HOST",
    default=_load_config().server.host,
    show_default=True,
    required=True,
)
def env_archive(
    project: str,
    environment: str,
    host: str,
):
    """
    Archive an environment on the server (but retain the configuration file locally).
    """
    environments = _api_request(
        "GET", host, "get_environments", params={"project": project}
    )
    environment_ids_by_name = {e["name"]: id for id, e in environments.items()}
    environment_id = environment_ids_by_name.get(environment)
    if not environment_id:
        raise click.BadOptionUsage("environment", "Not recognised")

    _api_request(
        "POST",
        host,
        "archive_environment",
        json={
            "projectId": project,
            "environmentId": environment_id,
        },
    )
    click.secho(f"Archived environment '{environment}'.", fg="green")


@cli.command("register")
@click.option(
    "-p",
    "--project",
    help="Project ID",
    envvar="COFLUX_PROJECT",
    default=_load_config().project,
    show_default=True,
    required=True,
)
@click.option(
    "-e",
    "--environment",
    help="Environment name",
    envvar="COFLUX_ENVIRONMENT",
    default=_load_config().environment,
    show_default=True,
    required=True,
)
@click.option(
    "-h",
    "--host",
    help="Host to connect to",
    envvar="COFLUX_HOST",
    default=_load_config().project,
    show_default=True,
    required=True,
)
@click.argument("module_name", nargs=-1)
def register(
    project: str,
    environment: str,
    host: str,
    module_name: tuple[str],
) -> None:
    """
    Register repositories with the server.

    Paths to scripts can be passed instead of module names.

    Options will be loaded from the configuration file, unless overridden as arguments (or environment variables).
    """
    if not module_name:
        raise click.ClickException("No module(s) specified.")
    targets = _load_repositories(list(module_name))
    _register_manifests(project, environment, host, targets)
    click.secho("Repository manifests registered.", fg="green")


@cli.command("agent")
@click.option(
    "-p",
    "--project",
    help="Project ID",
    envvar="COFLUX_PROJECT",
    default=_load_config().project,
    show_default=True,
    required=True,
)
@click.option(
    "-e",
    "--environment",
    help="Environment name",
    envvar="COFLUX_ENVIRONMENT",
    default=_load_config().environment,
    show_default=True,
    required=True,
)
@click.option(
    "-h",
    "--host",
    help="Host to connect to",
    envvar="COFLUX_HOST",
    default=_load_config().server.host,
    show_default=True,
    required=True,
)
@click.option(
    "--provides",
    help="Features that this agent provides (to be matched with features that tasks require)",
    multiple=True,
    envvar="COFLUX_PROVIDES",
    default=_encode_provides(_load_config().provides),
    show_default=True,
)
@click.option(
    "--launch",
    help="The launch ID",
    envvar="COFLUX_LAUNCH",
)
@click.option(
    "--concurrency",
    type=int,
    help="Limit on number of executions to process at once",
    default=_load_config().concurrency,
    show_default=True,
)
@click.option(
    "--watch",
    is_flag=True,
    default=False,
    help="Enable auto-reload when code changes",
)
@click.option(
    "--register",
    is_flag=True,
    default=False,
    help="Automatically register repositories",
)
@click.option(
    "--dev",
    is_flag=True,
    default=False,
    help="Enable development mode (implies `--watch` and `--register`)",
)
@click.argument("module_name", nargs=-1)
def agent(
    project: str,
    environment: str,
    host: str,
    provides: tuple[str] | None,
    launch: str | None,
    concurrency: int,
    watch: bool,
    register: bool,
    dev: bool,
    module_name: tuple[str],
) -> None:
    """
    Start an agent.

    Loads the specified modules as repositories. Paths to scripts can be passed instead of module names.

    Options will be loaded from the configuration file, unless overridden as arguments (or environment variables).
    """
    if not module_name:
        raise click.ClickException("No module(s) specified.")
    provides_ = _parse_provides(provides)
    config = _load_config()
    args = (*module_name,)
    kwargs = {
        "project": project,
        "environment": environment,
        "host": host,
        "provides": provides_,
        "serialiser_configs": config and config.serialisers,
        "blob_threshold": config and config.blobs and config.blobs.threshold,
        "blob_store_configs": config and config.blobs and config.blobs.stores,
        "concurrency": concurrency,
        "launch_id": launch,
        "register": register or dev,
    }
    if watch or dev:
        filter = watchfiles.PythonFilter()
        watchfiles.run_process(
            ".",
            target=_init,
            args=args,
            kwargs=kwargs,
            callback=_callback,
            watch_filter=filter,
        )
    else:
        _init(*args, **kwargs)


@cli.command("submit")
@click.option(
    "-p",
    "--project",
    help="Project ID",
    default=_load_config().project,
    show_default=True,
    required=True,
)
@click.option(
    "-e",
    "--environment",
    help="Environment name",
    default=_load_config().environment,
    show_default=True,
    required=True,
)
@click.option(
    "-h",
    "--host",
    help="Host to connect to",
    default=_load_config().server.host,
    show_default=True,
    required=True,
)
@click.argument("repository")
@click.argument("target")
@click.argument("argument", nargs=-1)
def submit(
    project: str,
    environment: str,
    host: str,
    repository: str,
    target: str,
    argument: tuple[str],
) -> None:
    """
    Submit a workflow to be run.
    """
    # TODO: support overriding options?
    workflow = _api_request(
        "GET",
        host,
        "get_workflow",
        params={
            "project": project,
            "environment": environment,
            "repository": repository,
            "target": target,
        },
    )
    execute_after = (
        int((time.time() + workflow["delay"]) * 1000) if workflow["delay"] else None
    )
    # TODO: handle response
    _api_request(
        "POST",
        host,
        "submit_workflow",
        json={
            "projectId": project,
            "environmentName": environment,
            "repository": repository,
            "target": target,
            "arguments": [["json", a] for a in argument],
            "waitFor": workflow["waitFor"],
            "cache": workflow["cache"],
            "defer": workflow["defer"],
            "executeAfter": execute_after,
            "retries": workflow["retries"],
            "requires": workflow["requires"],
        },
    )
    click.secho("Workflow submitted.", fg="green")
    # TODO: follow logs?
    # TODO: wait for result?


if __name__ == "__main__":
    cli()



import functools
import typing as t
import datetime as dt
import inspect
import re
import json

from . import context, models

TARGET_KEY = "_coflux_target"

T = t.TypeVar("T")
P = t.ParamSpec("P")


def _json_dumps(obj: t.Any) -> str:
    return json.dumps(obj, separators=(",", ":"))


def _build_parameter(parameter: inspect.Parameter) -> models.Parameter:
    return models.Parameter(
        parameter.name,
        # TODO: better serialisation?
        (
            str(parameter.annotation)
            if parameter.annotation != inspect.Parameter.empty
            else None
        ),
        (
            _json_dumps(parameter.default)
            if parameter.default != inspect.Parameter.empty
            else None
        ),
    )


def _parse_wait(
    wait: bool | t.Iterable[str] | str, parameters: list[models.Parameter]
) -> set[int]:
    if wait is True:
        return set(range(len(parameters)))
    if wait is False:
        return set()
    return set(_get_param_indexes(parameters, wait))


def _parse_cache(
    cache: bool | float | dt.timedelta,
    cache_params: t.Iterable[str] | str | None,
    cache_namespace: str | None,
    cache_version: str | None,
    parameters: list[models.Parameter],
) -> models.Cache | None:
    if not cache:
        return None
    return models.Cache(
        (
            True
            if cache_params is None
            else _get_param_indexes(parameters, cache_params)
        ),
        (
            int(cache * 1000)
            if isinstance(cache, (int, float)) and not isinstance(cache, bool)
            else (int(cache.total_seconds() * 1000) if isinstance(cache, dt.timedelta) else None)
        ),
        cache_namespace,
        cache_version,
    )


def _parse_retries(
    retries: int | tuple[int, int] | tuple[int, int, int]
) -> models.Retries | None:
    # TODO: parse string (e.g., '1h')
    match retries:
        case 0:
            return None
        case int(limit):
            return models.Retries(limit, 0, 0)
        case (limit, delay):
            return models.Retries(limit, delay, delay)
        case (limit, delay_min, delay_max):
            return models.Retries(limit, delay_min, delay_max)
        case other:
            raise ValueError(other)


def _parse_defer(
    defer: bool,
    defer_params: t.Iterable[str] | str | None,
    parameters: list[models.Parameter],
) -> models.Defer | None:
    if not defer:
        return None
    return models.Defer(
        (True if defer_params is None else _get_param_indexes(parameters, defer_params))
    )


def _parse_delay(delay: float | dt.timedelta) -> float:
    if isinstance(delay, dt.timedelta):
        return delay.total_seconds()
    return delay


def _parse_memo(
    memo: bool | t.Iterable[str] | str, parameters: list[models.Parameter]
) -> list[int] | bool:
    if isinstance(memo, bool):
        return memo
    return _get_param_indexes(parameters, memo)


def _build_definition(
    type: models.TargetType,
    fn: t.Callable,
    wait: bool | t.Iterable[str] | str,
    cache: bool | float | dt.timedelta,
    cache_params: t.Iterable[str] | str | None,
    cache_namespace: str | None,
    cache_version: str | None,
    retries: int | tuple[int, int] | tuple[int, int, int],
    defer: bool,
    defer_params: t.Iterable[str] | str | None,
    delay: float | dt.timedelta,
    memo: bool | t.Iterable[str] | str,
    requires: dict[str, str | bool | list[str]] | None,
    is_stub: bool,
) -> models.Target:
    parameters = inspect.signature(fn).parameters.values()
    for p in parameters:
        if p.kind != inspect.Parameter.POSITIONAL_OR_KEYWORD:
            raise Exception(f"Unsupported parameter type ({p.kind})")
    parameters_ = [_build_parameter(p) for p in parameters]
    return models.Target(
        type,
        parameters_,
        _parse_wait(wait, parameters_),
        _parse_cache(cache, cache_params, cache_namespace, cache_version, parameters_),
        _parse_defer(defer, defer_params, parameters_),
        _parse_delay(delay),
        _parse_retries(retries),
        _parse_memo(memo, parameters_),
        _parse_requires(requires),
        inspect.getdoc(fn),
        is_stub,
    )


class Target(t.Generic[P, T]):
    def __init__(
        self,
        fn: t.Callable[P, T],
        type: models.TargetType,
        *,
        repository: str | None = None,
        name: str | None = None,
        wait: bool | t.Iterable[str] | str = False,
        cache: bool | float | dt.timedelta = False,
        cache_params: t.Iterable[str] | str | None = None,
        cache_namespace: str | None = None,
        cache_version: str | None = None,
        retries: int | tuple[int, int] | tuple[int, int, int] = 0,
        defer: bool = False,
        defer_params: t.Iterable[str] | str | None = None,
        delay: float | dt.timedelta = 0,
        memo: bool | t.Iterable[str] | str = False,
        requires: dict[str, str | bool | list[str]] | None = None,
        is_stub: bool = False,
    ):
        self._fn = fn
        self._name = name or fn.__name__
        self._repository = repository or fn.__module__
        self._definition = _build_definition(
            type,
            fn,
            wait,
            cache,
            cache_params,
            cache_namespace,
            cache_version,
            retries,
            defer,
            defer_params,
            delay,
            memo,
            requires,
            is_stub,
        )
        functools.update_wrapper(self, fn)

    @property
    def name(self) -> str:
        return self._name

    @property
    def definition(self) -> models.Target:
        return self._definition

    @property
    def fn(self) -> t.Callable[P, T]:
        return self._fn

    def submit(self, *args: P.args, **kwargs: P.kwargs) -> models.Execution[T]:
        if kwargs:
            raise Exception("Keyword arguments aren't supported - pass positional arguments instead")
        try:
            return context.submit(
                self._definition.type,
                self._repository,
                self._name,
                args,
                wait_for=self._definition.wait_for,
                cache=self._definition.cache,
                retries=self._definition.retries,
                defer=self._definition.defer,
                delay=self._definition.delay,
                memo=self._definition.memo,
                requires=self._definition.requires,
            )
        except context.NotInContextException:
            result = self._fn(*args, **kwargs)
            return (
                models.Execution(lambda: result, lambda: None, None)
                if not isinstance(result, models.Execution)
                else result
            )

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.submit(*args, **kwargs).result()


def _get_param_indexes(
    parameters: list[models.Parameter],
    names: t.Iterable[str] | str,
) -> list[int]:
    if isinstance(names, str):
        names = re.split(r",\s*", names)
    indexes = []
    parameter_names = [p.name for p in parameters]
    for name in names:
        if name not in parameter_names:
            raise Exception(f"Unrecognised parameter in wait ({name})")
        indexes.append(parameter_names.index(name))
    return indexes


def _parse_require(value: str | bool | list[str]):
    if isinstance(value, bool):
        return ["true"] if value else ["false"]
    elif isinstance(value, str):
        return [value]
    else:
        return value


def _parse_requires(
    requires: dict[str, str | bool | list[str]] | None
) -> models.Requires | None:
    return {k: _parse_require(v) for k, v in requires.items()} if requires else None


def task(
    *,
    name: str | None = None,
    wait: bool | t.Iterable[str] | str = False,
    cache: bool | float | dt.timedelta = False,
    cache_params: t.Iterable[str] | str | None = None,
    cache_namespace: str | None = None,
    cache_version: str | None = None,
    retries: int | tuple[int, int] | tuple[int, int, int] = 0,
    defer: bool = False,
    defer_params: t.Iterable[str] | str | None = None,
    delay: float | dt.timedelta = 0,
    memo: bool | t.Iterable[str] = False,
    requires: dict[str, str | bool | list[str]] | None = None,
) -> t.Callable[[t.Callable[P, T]], Target[P, T]]:
    def decorator(fn: t.Callable[P, T]) -> Target[P, T]:
        return Target(
            fn,
            "task",
            name=name,
            wait=wait,
            cache=cache,
            cache_params=cache_params,
            cache_namespace=cache_namespace,
            cache_version=cache_version,
            retries=retries,
            defer=defer,
            defer_params=defer_params,
            delay=delay,
            memo=memo,
            requires=requires,
        )

    return decorator


def workflow(
    *,
    name: str | None = None,
    wait: bool | t.Iterable[str] | str = False,
    cache: bool | float | dt.timedelta = False,
    cache_params: t.Iterable[str] | str | None = None,
    cache_namespace: str | None = None,
    cache_version: str | None = None,
    retries: int | tuple[int, int] | tuple[int, int, int] = 0,
    defer: bool = False,
    defer_params: t.Iterable[str] | str | None = None,
    delay: float | dt.timedelta = 0,
    requires: dict[str, str | bool | list[str]] | None = None,
) -> t.Callable[[t.Callable[P, T]], Target[P, T]]:
    def decorator(fn: t.Callable[P, T]) -> Target[P, T]:
        return Target(
            fn,
            "workflow",
            name=name,
            wait=wait,
            cache=cache,
            cache_params=cache_params,
            cache_namespace=cache_namespace,
            cache_version=cache_version,
            retries=retries,
            defer=defer,
            defer_params=defer_params,
            delay=delay,
            requires=requires,
        )

    return decorator


def stub(
    repository: str,
    *,
    name: str | None = None,
    type: t.Literal["workflow", "task"] = "task",
    wait: bool | t.Iterable[str] | str = False,
    cache: bool | float | dt.timedelta = False,
    cache_params: t.Iterable[str] | str | None = None,
    cache_namespace: str | None = None,
    cache_version: str | None = None,
    retries: int | tuple[int, int] | tuple[int, int, int] = 0,
    defer: bool = False,
    defer_params: t.Iterable[str] | str | None = None,
    delay: float | dt.timedelta = 0,
    memo: bool | t.Iterable[str] = False,
) -> t.Callable[[t.Callable[P, T]], Target[P, T]]:
    def decorator(fn: t.Callable[P, T]) -> Target[P, T]:
        return Target(
            fn,
            type,
            repository=repository,
            name=name,
            wait=wait,
            cache=cache,
            cache_params=cache_params,
            cache_namespace=cache_namespace,
            cache_version=cache_version,
            retries=retries,
            defer=defer,
            defer_params=defer_params,
            delay=delay,
            memo=memo,
            is_stub=True,
        )

    return decorator


def sensor(
    *,
    name=None,
    requires: dict[str, str | bool | list[str]] | None = None,
) -> t.Callable[[t.Callable[P, None]], Target[P, None]]:
    def decorator(fn: t.Callable[P, None]) -> Target[P, None]:
        return Target(
            fn,
            "sensor",
            name=name,
            requires=requires,
        )

    return decorator



import importlib
import importlib.util
import types
from pathlib import Path


def load_module(module_name: str) -> types.ModuleType:
    path = Path(module_name)
    if path.is_file():
        spec = importlib.util.spec_from_file_location(module_name, path)
        assert spec and spec.loader
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    else:
        module = importlib.import_module(module_name)
    return module



import typing as t
from pathlib import Path

T = t.TypeVar("T")


TargetType = t.Literal["workflow", "task", "sensor"]


class Parameter(t.NamedTuple):
    name: str
    annotation: str | None
    default: str | None


class Cache(t.NamedTuple):
    params: list[int] | t.Literal[True]
    max_age: int | None
    namespace: str | None
    version: str | None


class Defer(t.NamedTuple):
    params: list[int] | t.Literal[True]


class Retries(t.NamedTuple):
    limit: int
    delay_min: int
    delay_max: int


Requires = dict[str, list[str]]


class Target(t.NamedTuple):
    type: TargetType
    parameters: list[Parameter]
    wait_for: set[int]
    cache: Cache | None
    defer: Defer | None
    delay: float
    retries: Retries | None
    memo: list[int] | bool
    requires: Requires | None
    instruction: str | None
    is_stub: bool


Metadata = dict[str, t.Any]

Reference = (
    tuple[t.Literal["execution"], int]
    | tuple[t.Literal["asset"], int]
    | tuple[t.Literal["fragment"], str, str, int, dict[str, t.Any]]
)

Value = t.Union[
    tuple[t.Literal["raw"], t.Any, list[Reference]],
    tuple[t.Literal["blob"], str, int, list[Reference]],
]

Result = t.Union[
    tuple[t.Literal["value"], Value],
    tuple[t.Literal["error"], str, str],
    tuple[t.Literal["abandoned"]],
    tuple[t.Literal["cancelled"]],
]


class Execution(t.Generic[T]):
    def __init__(
        self,
        resolve_fn: t.Callable[[], T],
        cancel_fn: t.Callable[[], None],
        id: int | None,
    ):
        self._resolve_fn = resolve_fn
        self._cancel_fn = cancel_fn
        self._id = id

    @property
    def id(self) -> int | None:
        return self._id

    def result(self) -> T:
        return self._resolve_fn()

    def cancel(self) -> None:
        self._cancel_fn()


class Asset:
    def __init__(
        self,
        restore_fn: t.Callable[[Path | str | None], Path],
        id: int,
    ):
        self._restore_fn = restore_fn
        self._id = id

    @property
    def id(self) -> int:
        return self._id

    def restore(self, *, to: Path | str | None = None) -> Path:
        return self._restore_fn(to)



import typing as t
import multiprocessing
import multiprocessing.connection
import threading
import time
import enum
import asyncio
import datetime as dt
import traceback
import sys
import tempfile
import os
import contextlib
import mimetypes
import zipfile
import itertools
import signal
from pathlib import Path
from contextvars import ContextVar
from contextlib import contextmanager

from . import server, blobs, models, serialisation, decorators, loader, config


_EXECUTION_THRESHOLD_S = 1.0
_AGENT_THRESHOLD_S = 5.0


T = t.TypeVar("T")

_channel_context: t.Optional["Channel"] = None
_timeout: ContextVar[float | None] = ContextVar("timeout", default=None)


class Timeout(Exception):
    pass


class Future(t.Generic[T]):
    def __init__(self):
        self._event = threading.Event()
        self._result: T | None = None
        self._error: t.Any | None = None

    def result(self, timeout: float | None = None):
        if not self._event.wait(timeout):
            raise Timeout()
        if self._error:
            raise Exception(self._error)
        else:
            return t.cast(T, self._result)

    def set_success(self, result):
        assert not self._result and not self._error
        self._result = result
        self._event.set()

    def set_error(self, error):
        assert not self._result and not self._error
        self._error = error
        self._event.set()


class ExecutionStatus(enum.Enum):
    STARTING = 0
    # TODO: preparing?
    EXECUTING = 1
    ABORTING = 2
    STOPPING = 3


class Error(t.NamedTuple):
    type: str
    message: str
    frames: list[tuple[str, int, str, str | None]]


class ExecutingNotification(t.NamedTuple):
    pass


class RecordResultRequest(t.NamedTuple):
    value: models.Value


class RecordErrorRequest(t.NamedTuple):
    error: Error


class SubmitExecutionRequest(t.NamedTuple):
    repository: str
    target: str
    type: models.TargetType
    arguments: list[models.Value]
    wait_for: set[int]
    cache: models.Cache | None
    defer: models.Defer | None
    memo: list[int] | bool
    execute_after: dt.datetime | None
    retries: models.Retries | None
    requires: models.Requires | None


class ResolveReferenceRequest(t.NamedTuple):
    execution_id: int


class PersistAssetRequest(t.NamedTuple):
    type: int
    path: str
    blob_key: str
    size: int
    metadata: dict[str, t.Any]


class ResolveAssetRequest(t.NamedTuple):
    asset_id: int


class SuspendRequest(t.NamedTuple):
    execute_after: dt.datetime | None
    execution_ids: list[int]


class CancelExecutionRequest(t.NamedTuple):
    execution_id: int


class RecordCheckpointRequest(t.NamedTuple):
    arguments: list[models.Value]


class LogMessageRequest(t.NamedTuple):
    level: int
    template: str | None
    values: dict[str, t.Any]
    timestamp: int


class RemoteException(Exception):
    def __init__(self, message, remote_type):
        super().__init__(message)
        self.remote_type = remote_type


def _exception_type(exception: Exception):
    if isinstance(exception, RemoteException):
        return exception.remote_type
    t = type(exception)
    if t.__module__ == "builtins":
        return t.__name__
    return f"{t.__module__}.{t.__name__}"


def _serialise_exception(exception: Exception) -> Error:
    type_ = _exception_type(exception)
    message = getattr(exception, "message", str(exception))
    frames = [
        (f.filename, f.lineno or 0, f.name, f.line)
        for f in traceback.extract_tb(exception.__traceback__)
    ]
    return Error(type_, message, frames)


class LineBuffer(t.IO[str]):
    def __init__(self, fn):
        self._fn = fn
        self._buffer = ""

    def write(self, content):
        self._buffer += content
        lines = self._buffer.split("\n")
        for line in lines[:-1]:
            self._fn(line)
        self._buffer = lines[-1]
        return len(content)

    def flush(self):
        if self._buffer:
            self._fn(self._buffer)
            self._buffer = ""


@contextlib.contextmanager
def capture_streams(*, stdout, stderr):
    original = (sys.stdout, sys.stderr)
    sys.stdout = LineBuffer(stdout)
    sys.stderr = LineBuffer(stderr)
    yield
    sys.stdout.flush()
    sys.stderr.flush()
    sys.stdout, sys.stderr = original


@contextlib.contextmanager
def working_directory():
    with tempfile.TemporaryDirectory() as temp_dir:
        path = Path(temp_dir)
        original = os.getcwd()
        os.chdir(path)
        yield path
        os.chdir(original)


def counter():
    count = itertools.count()
    return lambda: next(count)


class Channel:
    def __init__(
        self,
        execution_id: str,
        serialiser_configs: list[config.SerialiserConfig],
        blob_threshold: int,
        blob_store_configs: list[config.BlobStoreConfig],
        server_host: str,
        connection,
    ):
        self._execution_id = execution_id
        self._connection = connection
        self._request_id = counter()
        self._requests: dict[int, Future[t.Any]] = {}
        self._cache: dict[t.Any, Future[t.Any]] = {}
        self._running = True
        self._exit_stack = contextlib.ExitStack()
        self._blob_manager = blobs.Manager(blob_store_configs, server_host)
        self._serialisation_manager = serialisation.Manager(
            serialiser_configs, blob_threshold, self._blob_manager
        )

    def __enter__(self):
        self._directory = self._exit_stack.enter_context(working_directory())
        self._exit_stack.enter_context(self._blob_manager)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._exit_stack.close()

    def run(self):
        while self._running:
            if self._connection.poll(1):
                message = self._connection.recv()
                match message:
                    case ("success_result", (request_id, result)):
                        self._requests.pop(request_id).set_success(result)
                    case ("error_result", (request_id, error)):
                        self._requests.pop(request_id).set_error(error)
                    case other:
                        raise Exception(f"Received unhandled response: {other}")

    def _send(self, *message):
        self._connection.send(message)

    def _request(self, request, *, key=None, timeout=None):
        if key and key in self._cache:
            return self._cache[key].result(timeout)
        request_id = self._request_id()
        future = Future()
        self._requests[request_id] = future
        if key:
            self._cache[key] = future
        self._send("request", request_id, request)
        return future.result(timeout)

    def _notify(self, notification):
        self._send("notify", notification)

    def _record_result(self, data: t.Any):
        value = self._serialisation_manager.serialise(data)
        self._notify(RecordResultRequest(value))
        # TODO: wait for confirmation?
        self._running = False

    def _record_error(self, exception):
        self._notify(RecordErrorRequest(_serialise_exception(exception)))
        # TODO: wait for confirmation?
        self._running = False

    def submit_execution(
        self,
        type: models.TargetType,
        repository: str,
        target: str,
        arguments: tuple[t.Any, ...],
        *,
        wait_for: set[int] | None = None,
        cache: models.Cache | None = None,
        retries: models.Retries | None = None,
        defer: models.Defer | None = None,
        execute_after: dt.datetime | None = None,
        delay: float | dt.timedelta = 0,
        memo: list[int] | bool = False,
        requires: models.Requires | None = None,
    ) -> models.Execution[t.Any]:
        if delay:
            delay = (
                dt.timedelta(seconds=delay)
                if isinstance(delay, (int, float))
                else delay
            )
            execute_after = (execute_after or dt.datetime.now()) + delay
        # TODO: parallelise?
        serialised_arguments = [
            self._serialisation_manager.serialise(a) for a in arguments
        ]
        execution_id = self._request(
            SubmitExecutionRequest(
                repository,
                target,
                type,
                serialised_arguments,
                wait_for or set(),
                cache,
                defer,
                memo,
                execute_after,
                retries,
                requires,
            )
        )
        return models.Execution(
            lambda: self._resolve_reference(execution_id),
            lambda: self._cancel_execution(execution_id),
            execution_id,
        )

    @contextmanager
    def suspense(self, timeout: float | None):
        token = _timeout.set(timeout)
        try:
            yield
        finally:
            _timeout.reset(token)

    def suspend(self, delay: float | dt.timedelta | dt.datetime | None = None):
        execute_after = None
        if isinstance(delay, dt.datetime):
            execute_after = delay
        elif isinstance(delay, (int, float)):
            execute_after = dt.datetime.now() + dt.timedelta(seconds=delay)
        elif isinstance(delay, dt.timedelta):
            execute_after = dt.datetime.now() + delay
        self._notify(SuspendRequest(execute_after, []))

    def _resolve_arguments(
        self,
        arguments: list[models.Value],
    ) -> list[t.Any]:
        # TODO: parallelise
        return [
            self._serialisation_manager.deserialise(
                value,
                self._resolve_reference,
                self._cancel_execution,
                self._restore_asset,
            )
            for value in arguments
        ]

    def execute(self, target, arguments):
        resolved_arguments = self._resolve_arguments(arguments)
        self._notify(ExecutingNotification())
        try:
            with capture_streams(
                stdout=lambda m: self.log_message(1, m),
                stderr=lambda m: self.log_message(3, m),
            ):
                value = target.fn(*resolved_arguments)
        except KeyboardInterrupt:
            pass
        except Exception as e:
            self._record_error(e)
        else:
            self._record_result(value)

    def _deserialise_result(self, result: models.Result):
        match result:
            case ["value", value]:
                return self._serialisation_manager.deserialise(
                    value,
                    self._resolve_reference,
                    self._cancel_execution,
                    self._restore_asset,
                )
            case ["error", type_, message]:
                raise _build_exception(type_, message)
            case ["abandoned"]:
                raise Exception("abandoned")
            case ["cancelled"]:
                raise Exception("cancelled")
            case ["suspended"]:
                raise Exception("suspended")
            case result:
                raise Exception(f"unexpected result ({result})")

    def _resolve_reference(self, execution_id: int) -> t.Any:
        try:
            result = self._request(
                ResolveReferenceRequest(execution_id),
                key=("result", execution_id),
                timeout=_timeout.get(),
            )
            return self._deserialise_result(result)
        except Timeout:
            self._notify(SuspendRequest(None, [execution_id]))

    def _cancel_execution(self, execution_id: int) -> None:
        # TODO: wait for confirmation?
        self._notify(CancelExecutionRequest(execution_id))

    def record_checkpoint(self, arguments):
        serialised_arguments = [
            self._serialisation_manager.serialise(a) for a in arguments
        ]
        self._notify(RecordCheckpointRequest(serialised_arguments))

    def persist_asset(
        self, path: Path | str | None = None, *, match: str | None = None
    ) -> models.Asset:
        if isinstance(path, str):
            path = Path(path)
        path = path or self._directory
        if not path.exists():
            raise Exception(f"path '{path}' doesn't exist")
        path = path.resolve()
        if not path.is_relative_to(self._directory):
            raise Exception(f"path ({path}) not in execution directory")
        # TODO: zip all assets?
        path_str = str(path.relative_to(self._directory))
        if path.is_file():
            if match:
                raise Exception("match cannot be specified for file")
            asset_type = 0
            blob_key = self._blob_manager.upload(path)
            (mime_type, _) = mimetypes.guess_type(path)
            size = path.stat().st_size
            metadata = {"type": mime_type}
        elif path.is_dir():
            asset_type = 1
            sizes = []
            with tempfile.NamedTemporaryFile() as zip_file:
                zip_path = Path(zip_file.name)
                with zipfile.ZipFile(zip_path, "w") as zip:
                    for root, _, files in os.walk(path):
                        root = Path(root)
                        for file in files:
                            file_path = root.joinpath(file)
                            if not match or file_path.match(match):
                                zip.write(
                                    file_path, arcname=file_path.relative_to(path)
                                )
                                sizes.append(file_path.stat().st_size)
                blob_key = self._blob_manager.upload(zip_path)
            size = sum(sizes)
            metadata = {"count": len(sizes)}
        else:
            raise Exception(f"path ({path}) isn't a file or a directory")
        asset_id = self._request(
            PersistAssetRequest(asset_type, path_str, blob_key, size, metadata)
        )
        return models.Asset(lambda to: self._restore_asset(asset_id, to=to), asset_id)

    def _restore_asset(self, asset_id: int, to: Path | str | None = None) -> Path:
        if to:
            if isinstance(to, str):
                to = Path(to)
            if not to.is_absolute():
                to = self._directory.joinpath(to)
            if not to.is_relative_to(self._directory):
                raise Exception("asset must be restored to execution directory")
        # TODO: use timeout?
        asset_type, path_str, blob_key = self._request(
            ResolveAssetRequest(asset_id), key=("asset", asset_id)
        )
        target = to or self._directory.joinpath(path_str)
        target.parent.mkdir(parents=True, exist_ok=True)
        if asset_type == 0:
            self._blob_manager.download(blob_key, target)
        elif asset_type == 1:
            target.mkdir()
            with tempfile.NamedTemporaryFile() as zip_file:
                zip_path = Path(zip_file.name)
                self._blob_manager.download(blob_key, zip_path)
                with zipfile.ZipFile(zip_path, "r") as zip:
                    zip.extractall(target)
        else:
            raise Exception(f"unrecognised asset type ({asset_type})")
        return target

    def log_message(self, level, template: str | None, **kwargs):
        timestamp = time.time() * 1000

        values = {
            key: self._serialisation_manager.serialise(value)
            for key, value in kwargs.items()
        }
        self._notify(
            LogMessageRequest(
                level,
                str(template) if template is not None else None,
                values,
                int(timestamp),
            )
        )


def get_channel() -> Channel | None:
    return _channel_context


def _build_exception(type_, message):
    # TODO: better way to re-create exception?
    parts = type_.rsplit(type_, maxsplit=1)
    module = "builtins" if len(parts) == 1 else parts[0]
    name = parts[-1]
    if module in sys.modules:
        class_ = sys.modules[module].get(name)
        if class_:
            return class_(message)
    return RemoteException(message, type_)


def _execute(
    module_name: str,
    target_name: str,
    arguments: list[models.Value],
    execution_id: str,
    serialiser_configs: list[config.SerialiserConfig],
    blob_threshold: int,
    blob_store_configs: list[config.BlobStoreConfig],
    server_host: str,
    conn,
):
    global _channel_context
    module = loader.load_module(module_name)
    with Channel(
        execution_id,
        serialiser_configs,
        blob_threshold,
        blob_store_configs,
        server_host,
        conn,
    ) as channel:
        threading.Thread(target=channel.run).start()
        _channel_context = channel
        try:
            target = getattr(module, target_name)
            if not isinstance(target, decorators.Target) or target.definition.is_stub:
                raise Exception("not valid target")
            channel.execute(target, arguments)
        finally:
            _channel_context = None


def _json_safe_reference(reference: models.Reference) -> t.Any:
    match reference:
        case ("execution", execution_id):
            return ["execution", execution_id]
        case ("asset", asset_id):
            return ["asset", asset_id]
        case ("fragment", serialiser, blob_key, size, metadata):
            return ["fragment", serialiser, blob_key, size, metadata]
        case other:
            raise Exception(f"unhandled reference type ({other})")


def _json_safe_references(references: list[models.Reference]) -> list[t.Any]:
    return [_json_safe_reference(r) for r in references]


def _json_safe_value(value: models.Value):
    # TODO: tidy
    match value:
        case ("raw", content, references):
            return ["raw", content, _json_safe_references(references)]
        case ("blob", key, size, references):
            return ["blob", key, size, _json_safe_references(references)]


def _json_safe_arguments(arguments: list[models.Value]):
    return [_json_safe_value(v) for v in arguments]


def _parse_reference(reference) -> models.Reference:
    match reference:
        case ["execution", execution_id]:
            return ("execution", execution_id)
        case ["asset", asset_id]:
            return ("asset", asset_id)
        case ["fragment", serialiser, blob_key, size, metadata]:
            return ("fragment", serialiser, blob_key, size, metadata)
        case other:
            raise Exception(f"unrecognised reference: {other}")


def _parse_references(references) -> list[models.Reference]:
    return [_parse_reference(r) for r in references]


def _parse_value(value: t.Any) -> models.Value:
    match value:
        case ["raw", content, references]:
            return ("raw", content, _parse_references(references))
        case ["blob", key, size, references]:
            return ("blob", key, size, _parse_references(references))
        case other:
            raise Exception(f"unrecognised value: {other}")


def _parse_result(result: t.Any) -> models.Result:
    match result:
        case ["error", type_, message]:
            return ("error", type_, message)
        case ["value", value]:
            return ("value", _parse_value(value))
        case ["abandoned"]:
            return ("abandoned",)
        case ["cancelled"]:
            return ("cancelled",)
        case other:
            raise Exception(f"unrecognised result: {other}")


class Execution:
    def __init__(
        self,
        execution_id: str,
        module: str,
        target: str,
        arguments: list[t.Any],
        serialiser_configs: list[config.SerialiserConfig],
        blob_threshold: int,
        blob_store_configs: list[config.BlobStoreConfig],
        server_host: str,
        server_connection: server.Connection,
        loop: asyncio.AbstractEventLoop,
    ):
        self._id = execution_id
        self._server = server_connection
        self._loop = loop
        self._timestamp = time.time()  # TODO: better name
        self._status = ExecutionStatus.STARTING
        mp_context = multiprocessing.get_context("spawn")
        parent_conn, child_conn = mp_context.Pipe()
        self._connection = parent_conn
        self._process = mp_context.Process(
            target=_execute,
            args=(
                module,
                target,
                arguments,
                execution_id,
                serialiser_configs,
                blob_threshold,
                blob_store_configs,
                server_host,
                child_conn,
            ),
            name=f"Execution-{execution_id}",
        )

    @property
    def id(self):
        return self._id

    @property
    def status(self):
        return self._status

    @property
    def timestamp(self):
        return self._timestamp

    def touch(self, timestamp):
        self._timestamp = timestamp

    @property
    def sentinel(self):
        return self._process.sentinel

    def interrupt(self):
        self._status = ExecutionStatus.ABORTING
        pid = self._process.pid
        if pid:
            os.kill(pid, signal.SIGINT)

    def kill(self):
        self._status = ExecutionStatus.ABORTING
        self._process.kill()

    def join(self, timeout):
        self._process.join(timeout)
        return self._process.exitcode

    def _server_notify(self, request, params):
        coro = self._server.notify(request, params)
        future = asyncio.run_coroutine_threadsafe(coro, self._loop)
        future.result()

    def _server_request(self, request, params, request_id, success_parser=None):
        success_parser = success_parser or (lambda x: x)
        coro = self._server.request(
            request,
            params,
            server.Callbacks(
                on_success=lambda result: self._try_send(
                    "success_result", (request_id, success_parser(result))
                ),
                on_error=lambda error: self._try_send(
                    "error_result", (request_id, error)
                ),
            ),
        )
        future = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return future.result()

    def _try_send(
        self,
        type: t.Literal["success_result", "error_result"],
        value: t.Any,
    ):
        try:
            self._connection.send((type, value))
        except BrokenPipeError:
            pass

    def _handle_notify(self, message):
        match message:
            case ExecutingNotification():
                self._status = ExecutionStatus.EXECUTING
            case RecordResultRequest(value):
                self._status = ExecutionStatus.STOPPING
                self._server_notify(
                    "put_result",
                    (self._id, _json_safe_value(value)),
                )
                self._process.join()
            case RecordErrorRequest(error):
                self._status = ExecutionStatus.STOPPING
                self._server_notify("put_error", (self._id, error))
                self._process.join()
            case SuspendRequest(execute_after, execution_ids):
                self._status = ExecutionStatus.STOPPING
                execute_after_ms = execute_after and int(
                    execute_after.timestamp() * 1000
                )
                self._server_notify(
                    "suspend", (self._id, execute_after_ms, execution_ids)
                )
                self._process.join()
            case CancelExecutionRequest(execution_id):
                # TODO: pass reference to self?
                self._server_notify("cancel", (execution_id,))
            case RecordCheckpointRequest(arguments):
                self._server_notify(
                    "record_checkpoint",
                    (self._id, _json_safe_arguments(arguments)),
                )
            case LogMessageRequest(level, template, values, timestamp):
                self._server_notify(
                    "log_messages", ([self._id, timestamp, level, template, values],)
                )
            case other:
                raise Exception(f"Received unhandled notify: {other!r}")

    def _handle_request(self, request_id, request):
        match request:
            case SubmitExecutionRequest(
                repository,
                target,
                type,
                arguments,
                wait_for,
                cache,
                defer,
                memo,
                execute_after,
                retries,
                requires,
            ):
                execute_after_ms = execute_after and int(
                    execute_after.timestamp() * 1000
                )
                self._server_request(
                    "submit",
                    (
                        repository,
                        target,
                        type,
                        _json_safe_arguments(arguments),
                        self._id,
                        list(wait_for),
                        cache and cache._asdict(),
                        defer and defer._asdict(),
                        memo,
                        execute_after_ms,
                        retries and retries._asdict(),
                        requires,
                    ),
                    request_id,
                )
            case ResolveReferenceRequest(execution_id):
                # TODO: set (and unset) state on Execution to indicate waiting?
                self._server_request(
                    "get_result", (execution_id, self._id), request_id, _parse_result
                )
            case PersistAssetRequest(type, path, blob_key, size, metadata):
                self._server_request(
                    "put_asset",
                    (self._id, type, path, blob_key, size, metadata),
                    request_id,
                )
            case ResolveAssetRequest(asset_id):
                self._server_request("get_asset", (asset_id, self._id), request_id)
            case other:
                raise Exception(f"Received unhandled request: {other!r}")

    def run(self):
        self._process.start()
        while self._process.is_alive():
            if self._connection.poll(1):
                try:
                    message = self._connection.recv()
                except EOFError:
                    pass
                else:
                    match message:
                        case ("notify", notify):
                            self._handle_notify(notify)
                        case ("request", request_id, request):
                            self._handle_request(request_id, request)
                        case other:
                            raise Exception(f"Received unhandled message: {other!r}")


def _wait_processes(sentinels: t.Iterable[int], until: float) -> set[int]:
    remaining = set(sentinels)
    while sentinels and time.time() < until:
        ready = multiprocessing.connection.wait(remaining, until - time.time())
        for sentinel in ready:
            assert isinstance(sentinel, int)
            remaining.remove(sentinel)
    return remaining


class Manager:
    def __init__(
        self,
        connection: server.Connection,
        serialiser_configs: list[config.SerialiserConfig],
        blob_threshold: int,
        blob_store_configs: list[config.BlobStoreConfig],
    ):
        self._connection = connection
        self._serialiser_configs = serialiser_configs
        self._blob_threshold = blob_threshold
        self._blob_store_configs = blob_store_configs
        self._executions: dict[str, Execution] = {}
        self._last_heartbeat_sent = None

    async def _declare_targets(
        self, targets: dict[str, dict[models.TargetType, list[str]]]
    ) -> None:
        await self._connection.notify("declare_targets", (targets,))

    async def _send_heartbeats(self) -> t.NoReturn:
        while True:
            now = time.time()
            executions = [e for e in self._executions.values()]
            if self._should_send_heartbeats(executions, _AGENT_THRESHOLD_S, now):
                heartbeats = {e.id: e.status.value for e in executions}
                await self._connection.notify("record_heartbeats", (heartbeats,))
                self._last_heartbeat_sent = now
                for execution in executions:
                    execution.touch(now)
            elapsed = time.time() - now
            await asyncio.sleep(_EXECUTION_THRESHOLD_S - elapsed)

    def _should_send_heartbeats(
        self, executions: list[Execution], threshold_s: float, now: float
    ) -> bool:
        return (
            any(executions)
            or not self._last_heartbeat_sent
            or (now - self._last_heartbeat_sent) > threshold_s
        )

    async def run(self, targets: dict[str, dict[models.TargetType, list[str]]]) -> None:
        # TODO: only declare targets once when starting session? (not on reconnect)
        await self._declare_targets(targets)
        await self._send_heartbeats()

    def execute(
        self,
        execution_id: str,
        module: str,
        target: str,
        arguments: list[models.Value],
        server_host: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        if execution_id in self._executions:
            raise Exception(f"Execution ({execution_id}) already running")
        execution = Execution(
            execution_id,
            module,
            target,
            arguments,
            self._serialiser_configs,
            self._blob_threshold,
            self._blob_store_configs,
            server_host,
            self._connection,
            loop,
        )
        threading.Thread(
            target=self._run_execution,
            args=(execution, loop),
        ).start()

    def _run_execution(
        self, execution: Execution, loop: asyncio.AbstractEventLoop
    ) -> None:
        self._executions[execution.id] = execution
        try:
            execution.run()
        finally:
            del self._executions[execution.id]
            coro = self._connection.notify("notify_terminated", ([execution.id],))
            future = asyncio.run_coroutine_threadsafe(coro, loop)
            future.result()

    def abort(self, execution_id: str, timeout: int = 5) -> bool:
        execution = self._executions.get(execution_id)
        if not execution:
            return False
        execution.interrupt()
        if execution.join(timeout) is not None:
            print(
                f"Execution ({execution_id}) hasn't exited within timeout {timeout}. Killing..."
            )
            execution.kill()
        return True

    def abort_all(self, timeout: int = 5):
        until = time.time() + timeout
        executions = {e.sentinel: e for e in self._executions.values()}
        for execution in executions.values():
            execution.interrupt()
        remaining = _wait_processes(executions.keys(), until)
        if remaining:
            remaining_ids = [str(executions[s].id) for s in remaining]
            print(
                f"Executions ({', '.join(remaining_ids)}) haven't exited within timeout {timeout}. Killing..."
            )
            for execution in executions.values():
                execution.kill()



import httpx
import hashlib
import typing as t
import io
import abc
from pathlib import Path

from . import config

try:
    import boto3
    import botocore
except ImportError:
    boto3 = None
    botocore = None


def _hash_file(buffer: t.BinaryIO):
    buffer.seek(0)
    hash = hashlib.sha256()
    for chunk in iter(lambda: buffer.read(4096), b""):
        hash.update(chunk)
    return hash.hexdigest()


class Store(abc.ABC):
    @abc.abstractmethod
    def __enter__(self):
        return self

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @abc.abstractmethod
    def get(self, key: str) -> io.BytesIO | None:
        raise NotImplementedError

    @abc.abstractmethod
    def put(self, buffer: t.BinaryIO) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def download(self, key: str, path: Path) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def upload(self, path: Path) -> str:
        raise NotImplementedError


class HttpStore(Store):
    def __init__(self, protocol: t.Literal["http", "https"], host: str):
        self._protocol = protocol
        self._host = host

    def __enter__(self):
        self._client = httpx.Client(timeout=10)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._client.close()

    def _url(self, key: str) -> str:
        return f"{self._protocol}://{self._host}/blobs/{key}"

    def _exists(self, key: str) -> bool:
        return self._client.head(self._url(key)).status_code == 200

    def get(self, key: str) -> io.BytesIO | None:
        response = self._client.get(self._url(key))
        if response.status_code == 404:
            return None
        response.raise_for_status()
        return io.BytesIO(response.content)

    def put(self, buffer: t.BinaryIO) -> str:
        assert buffer.seekable()
        key = _hash_file(buffer)
        if not self._exists(key):
            buffer.seek(0)
            self._client.put(self._url(key), content=buffer).raise_for_status()
        return key

    def download(self, key: str, path: Path) -> bool:
        with self._client.stream("GET", self._url(key)) as response:
            if response.status_code == 404:
                return False
            response.raise_for_status()
            with path.open("wb") as file:
                for chunk in response.iter_bytes():
                    file.write(chunk)
                return True

    def upload(self, path: Path) -> str:
        with open(path, "rb") as file:
            return self.put(file)


class S3Store(Store):
    def __init__(self, bucket_name: str, prefix: str | None, region: str | None):
        self._bucket_name = bucket_name
        self._prefix = prefix.strip("/") if prefix else ""
        self._region = region

    def __enter__(self):
        if not boto3:
            raise Exception("Missing 'boto3' dependency")
        self._s3 = boto3.client("s3", region_name=self._region)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _key(self, key: str) -> str:
        prefix = f"{self._prefix}/" if self._prefix else ""
        return f"{prefix}{key[0:2]}/{key[2:4]}/{key[4:]}"

    def _exists(self, key: str) -> bool:
        assert botocore
        try:
            self._s3.head_object(Bucket=self._bucket_name, Key=self._key(key))
            return True
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise

    def get(self, key: str) -> io.BytesIO | None:
        assert botocore
        try:
            response = self._s3.get_object(Bucket=self._bucket_name, Key=self._key(key))
            return io.BytesIO(response["Body"].read())
        except botocore.exceptions.NoSuchKey:
            return None

    def put(self, buffer: t.BinaryIO) -> str:
        assert buffer.seekable()
        key = _hash_file(buffer)
        if not self._exists(key):
            buffer.seek(0)
            self._s3.put_object(
                Bucket=self._bucket_name,
                Key=self._key(key),
                Body=buffer,
            )
        return key

    def download(self, key: str, path: Path) -> bool:
        assert botocore
        try:
            with open(path, "wb") as file:
                self._s3.download_fileobj(self._bucket_name, self._key(key), file)
                return True
        except botocore.exceptions.NoSuchKey:
            return False

    def upload(self, path: Path) -> str:
        with open(path, "rb") as file:
            return self.put(file)


def _create(config_: config.BlobStoreConfig, server_host: str):
    if config_.type == "http":
        return HttpStore(config_.protocol, config_.host or server_host)
    elif config_.type == "s3":
        return S3Store(config_.bucket, config_.prefix, config_.region)
    else:
        raise ValueError("unrecognised blob store config")


class Manager:
    def __init__(self, store_configs: list[config.BlobStoreConfig], server_host: str):
        self._stores = [_create(c, server_host) for c in store_configs]

    def __enter__(self):
        # TODO: ?
        for store in self._stores:
            store.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def get(self, key: str) -> io.BytesIO:
        for store in self._stores:
            result = store.get(key)
            if result is not None:
                return result
        raise Exception(f"blob not found ({key})")

    def put(self, buffer: t.BinaryIO) -> str:
        return self._stores[0].put(buffer)

    def download(self, key: str, path: Path) -> None:
        for store in self._stores:
            if store.download(key, path):
                return
        raise Exception(f"blob not found ({key})")

    def upload(self, path: Path) -> str:
        return self._stores[0].upload(path)



import typing as t
import json
import pickle
import abc
import io
import pydantic
import importlib
import collections
from pathlib import Path

try:
    import pandas
except ImportError:
    pandas = None

from . import blobs, models, config

T = t.TypeVar("T")


def _json_dumps(obj: t.Any) -> str:
    return json.dumps(obj, separators=(",", ":"))


class Serialiser(abc.ABC):
    @abc.abstractmethod
    def serialise(self, value: t.Any) -> tuple[str, io.BytesIO, dict[str, t.Any]] | None:
        raise NotImplementedError

    @abc.abstractmethod
    def deserialise(self, format: str, buffer: io.BytesIO, metadata: dict[str, t.Any]) -> t.Any | None:
        raise NotImplementedError


class PickleSerialiser(Serialiser):
    def serialise(self, value: t.Any) -> tuple[str, io.BytesIO, dict[str, t.Any]] | None:
        try:
            buffer = io.BytesIO()
            pickle.dump(value, buffer)
            return "pickle", buffer, {"type": str(type(value))}
        except pickle.PicklingError:
            return None

    def deserialise(self, format: str, buffer: io.BytesIO, metadata: dict[str, t.Any]) -> t.Any | None:
        if format != "pickle":
            return None
        return pickle.loads(buffer.getbuffer())


class PydanticSerialiser(Serialiser):
    def serialise(self, value: t.Any) -> tuple[str, io.BytesIO, dict[str, t.Any]] | None:
        if not isinstance(value, pydantic.BaseModel):
            return None
        buffer = io.BytesIO()
        json_data = value.model_dump_json().encode()
        buffer.write(json_data)
        model_class = value.__class__
        model = f"{model_class.__module__}.{model_class.__name__}"
        return "json", buffer, {"model": model}

    def deserialise(self, format: str, buffer: io.BytesIO, metadata: dict[str, t.Any]) -> t.Any | None:
        if format != "json" or "model" not in metadata:
            return None
        json_data = buffer.read().decode()
        model = metadata["model"]
        module_name, class_name = model.rsplit(".", 1)
        # TODO: support configuring which modules can be loaded?
        module = importlib.import_module(module_name)
        model_class = getattr(module, class_name)
        if not model_class:
            raise ValueError(f"Model '{model}' not found")
        if not issubclass(model_class, pydantic.BaseModel):
            raise ValueError(f"Model '{model}' isn't a subclass of BaseModel")
        return model_class.model_validate_json(json_data)


class PandasSerialiser(Serialiser):
    def serialise(self, value: t.Any) -> tuple[str, io.BytesIO, dict[str, t.Any]] | None:
        # TODO: support series
        if pandas and isinstance(value, pandas.DataFrame):
            buffer = io.BytesIO()
            value.to_parquet(buffer)
            return "parquet", buffer, {"shape": value.shape}

    def deserialise(self, format: str, buffer: io.BytesIO, metadata: dict[str, t.Any]) -> t.Any | None:
        if format != "parquet":
            return None
        if not pandas:
            raise Exception("Pandas dependency not available")
        # TODO: check metadata["content_type"]
        return pandas.read_parquet(buffer)


def _create(config_: config.SerialiserConfig):
    if config_.type == "pickle":
        return PickleSerialiser()
    elif config_.type == "pydantic":
        return PydanticSerialiser()
    elif config_.type == "pandas":
        return PandasSerialiser()
    else:
        raise ValueError("unrecognised serialiser config")


class Manager:
    def __init__(
        self,
        serialiser_configs: list[config.SerialiserConfig],
        blob_threshold: int,
        blob_manager: blobs.Manager,
    ):
        self._serialisers = [_create(c) for c in serialiser_configs]
        self._blob_threshold = blob_threshold
        self._blob_manager = blob_manager

    def serialise(self, value: t.Any) -> models.Value:
        references: list[models.Reference] = []

        def _serialise(value: t.Any) -> t.Any:
            if value is None or isinstance(value, (str, bool, int, float)):
                return value
            elif isinstance(value, list):
                return [_serialise(v) for v in value]
            elif isinstance(value, dict):
                items = (
                    value.items()
                    if isinstance(value, collections.OrderedDict)
                    else sorted(value.items(), key=lambda kv: repr(kv[0]))
                )
                return {
                    "type": "dict",
                    "items": [_serialise(x) for kv in items for x in kv],
                }
            elif isinstance(value, set):
                return {
                    "type": "set",
                    "items": [_serialise(x) for x in sorted(value)],
                }
            elif isinstance(value, models.Execution):
                # TODO: better handle id being none
                assert value.id is not None
                references.append(("execution", value.id))
                return {"type": "ref", "index": len(references) - 1}
            elif isinstance(value, models.Asset):
                references.append(("asset", value.id))
                return {"type": "ref", "index": len(references) - 1}
            elif isinstance(value, tuple):
                # TODO: include name
                return {"type": "tuple", "items": [_serialise(x) for x in value]}
            else:
                for serialiser in self._serialisers:
                    result = serialiser.serialise(value)
                    if result is not None:
                        format, buffer, metadata = result
                        blob_key = self._blob_manager.put(buffer)
                        size = buffer.getbuffer().nbytes
                        references.append(
                            ("fragment", format, blob_key, size, metadata)
                        )
                        return {"type": "ref", "index": len(references) - 1}
                raise Exception(f"no serialiser for type '{type(value)}'")

        data = _serialise(value)
        encoded_data = _json_dumps(data).encode()
        size = len(encoded_data)
        if size > self._blob_threshold:
            buffer = io.BytesIO(encoded_data)
            blob_key = self._blob_manager.put(buffer)
            return ("blob", blob_key, size, references)
        else:
            return ("raw", data, references)

    def _get_value_data(
        self, value: models.Value
    ) -> tuple[t.Any, list[models.Reference]]:
        match value:
            case ("blob", key, _, references):
                result = self._blob_manager.get(key)
                return json.load(result), references
            case ("raw", data, references):
                return data, references

    def deserialise(
        self,
        value: models.Value,
        resolve_fn: t.Callable[[int], t.Any],
        cancel_fn: t.Callable[[int], None],
        restore_fn: t.Callable[[int, Path | str | None], Path],
    ) -> t.Any:
        data, references = self._get_value_data(value)

        def _deserialise(data: t.Any):
            if data is None or isinstance(data, (str, bool, int, float)):
                return data
            elif isinstance(data, list):
                return [_deserialise(v) for v in data]
            elif isinstance(data, dict):
                match data["type"]:
                    case "dict":
                        pairs = zip(data["items"][::2], data["items"][1::2])
                        return {_deserialise(k): _deserialise(v) for k, v in pairs}
                    case "set":
                        return {_deserialise(x) for x in data["items"]}
                    case "tuple":
                        return tuple(_deserialise(x) for x in data["items"])
                    case "ref":
                        reference = references[data["index"]]
                        match reference:
                            case ("execution", execution_id):
                                return models.Execution(
                                    lambda: resolve_fn(execution_id),
                                    lambda: cancel_fn(execution_id),
                                    execution_id,
                                )
                            case ("asset", asset_id):
                                return models.Asset(
                                    lambda to: restore_fn(asset_id, to), asset_id
                                )
                            case ("fragment", format, blob_key, _size, metadata):
                                data = self._blob_manager.get(blob_key)
                                for serialiser in self._serialisers:
                                    result = serialiser.deserialise(format, data, metadata)
                                    if result is not None:
                                        return result
                                raise Exception(f"Couldn't deserialise fragment ({format})")
                    case other:
                        raise Exception(f"unhandled data type ({other})")
            else:
                raise Exception(f"unhandled data type ({type(data)})")

        return _deserialise(data)



from .decorators import workflow, task, stub, sensor
from .context import (
    checkpoint,
    suspense,
    suspend,
    persist_asset,
    log_debug,
    log_info,
    log_warning,
    log_error,
)
from .models import Execution, Asset
from .agent import Agent

__all__ = [
    "workflow",
    "task",
    "stub",
    "sensor",
    "checkpoint",
    "suspense",
    "suspend",
    "log_debug",
    "log_info",
    "log_warning",
    "log_error",
    "persist_asset",
    "Execution",
    "Asset",
    "Agent",
]



import asyncio
import random
import typing as t
import urllib.parse
import websockets
import traceback

from . import server, execution, models, config


def _parse_reference(reference: t.Any) -> models.Reference:
    match reference:
        case ["execution", execution_id]:
            return ("execution", execution_id)
        case ["asset", asset_id]:
            return ("asset", asset_id)
        case ["fragment", serialiser, blob_key, size, metadata]:
            return ("fragment", serialiser, blob_key, size, metadata)
        case other:
            raise Exception(f"unexpected reference: {other}")


def _parse_references(references: list[t.Any]) -> list[models.Reference]:
    return [_parse_reference(r) for r in references]


# TODO: remove duplication in execution.py
def _parse_value(value: list) -> models.Value:
    match value:
        case ["raw", content, references]:
            return ("raw", content, _parse_references(references))
        case ["blob", key, size, references]:
            return ("blob", key, size, _parse_references(references))
    raise Exception(f"unexpected value: {value}")


def _encode_tags(provides: dict[str, list[str]]) -> str:
    return ";".join(f"{k}:{v}" for k, vs in provides.items() for v in vs)


class Agent:
    def __init__(
        self,
        project_id: str,
        environment_name: str,
        server_host: str,
        provides: dict[str, list[str]],
        serialiser_configs: list[config.SerialiserConfig],
        blob_threshold: int,
        blob_store_configs: list[config.BlobStoreConfig],
        concurrency: int,
        launch_id: str | None,
        targets: dict[str, dict[str, tuple[models.Target, t.Callable]]],
    ):
        self._project_id = project_id
        self._environment_name = environment_name
        self._launch_id = launch_id
        self._server_host = server_host
        self._provides = provides
        self._concurrency = concurrency
        self._targets = targets
        self._connection = server.Connection(
            {"execute": self._handle_execute, "abort": self._handle_abort}
        )
        self._execution_manager = execution.Manager(
            self._connection, serialiser_configs, blob_threshold, blob_store_configs
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._execution_manager.abort_all()

    async def _handle_execute(self, *args) -> None:
        (execution_id, repository, target_name, arguments) = args
        print(f"Handling execute '{target_name}' ({execution_id})...")
        target = self._targets[repository][target_name][1].__name__
        arguments = [_parse_value(a) for a in arguments]
        loop = asyncio.get_running_loop()
        self._execution_manager.execute(
            execution_id, repository, target, arguments, self._server_host, loop
        )

    async def _handle_abort(self, *args) -> None:
        (execution_id,) = args
        print(f"Aborting execution ({execution_id})...")
        if not self._execution_manager.abort(execution_id):
            print(f"Ignored abort for unrecognised execution ({execution_id}).")

    def _url(self, scheme: str, path: str, params: dict[str, str]) -> str:
        params_ = {k: v for k, v in params.items() if v is not None} if params else None
        query_string = f"?{urllib.parse.urlencode(params_)}" if params_ else ""
        return f"{scheme}://{self._server_host}/{path}{query_string}"

    def _params(self):
        params = {
            "project": self._project_id,
            "environment": self._environment_name,
        }
        if self._connection.session_id:
            params["session"] = self._connection.session_id
        elif self._launch_id:
            params["launch"] = self._launch_id
        else:
            if self._provides:
                params["provides"] = _encode_tags(self._provides)
            if self._concurrency:
                params["concurrency"] = str(self._concurrency)
        return params

    async def run(self) -> None:
        while True:
            print(
                f"Connecting ({self._server_host}, {self._project_id}, {self._environment_name})..."
            )
            url = self._url("ws", "agent", self._params())
            try:
                async with websockets.connect(url) as websocket:
                    print("Connected.")
                    targets: dict[str, dict[models.TargetType, list[str]]] = {}
                    for repository, repository_targets in self._targets.items():
                        for target_name, (target, _) in repository_targets.items():
                            targets.setdefault(repository, {}).setdefault(
                                target.type, []
                            ).append(target_name)
                    coros = [
                        asyncio.create_task(self._connection.run(websocket)),
                        asyncio.create_task(self._execution_manager.run(targets)),
                    ]
                    done, pending = await asyncio.wait(
                        coros, return_when=asyncio.FIRST_COMPLETED
                    )
                    for task in pending:
                        task.cancel()
                    for task in done:
                        task.result()
            except websockets.ConnectionClosedError as e:
                reason = e.rcvd.reason if e.rcvd else None
                if reason == "project_not_found":
                    print("Project not found")
                    return
                elif reason == "environment_not_found":
                    print("Environment not found")
                    return
                elif reason == "session_invalid":
                    print("Session expired. Resetting and reconnecting...")
                    self._connection.reset()
                    self._execution_manager.abort_all()
                else:
                    delay = 1 + 3 * random.random()  # TODO: exponential backoff
                    print(f"Disconnected (reconnecting in {delay:.1f} seconds).")
                    await asyncio.sleep(delay)
            except OSError:
                traceback.print_exc()
                delay = 1 + 3 * random.random()  # TODO: exponential backoff
                print(f"Can't connect (retrying in {delay:.1f} seconds).")
                await asyncio.sleep(delay)
